#include <common.h>
#include <malloc.h>


#define MAX_I2C_RETRYS	    10
#define I2C_DELAY	    1000  /* Should be at least the # of MHz of Tclk */
#undef	DEBUG_I2C

#ifdef DEBUG_I2C
#define DP(x) x
#else
#define DP(x)
#endif

#define BIT2            0x00000004
#define BIT3            0x00000008

#define REG_BASE			0xA8000000

#define RESET_REG_BITS(regOffset,bits) *(unsigned int*)(REG_BASE | regOffset) &= ~((unsigned int)bits)

#define CMIC_CONFIG			0x10C

#define I2C_SLAVE_ADDRESS		0x120
#define I2C_DATA			0x124
#define I2C_CONTROL			0x128
#define I2C_STATUS_BAUD_RATE		0x12C
#define I2C_SOFT_RESET			0x13C

#define REG_WRITE(r, v) *(volatile int *)((u32)REG_BASE + r) = v
#define REG_READ(r, v) *(int *)v = *(volatile int *)((u32)REG_BASE + r)

/* Assuming that there is only one master on the bus (us) */

void i2c_init(int speed, int slaveaddr)
{
	unsigned int n, m, freq, margin, power;
	unsigned int actualFreq, actualN=0, actualM=0;
	unsigned int control, status;
	unsigned int minMargin = 0xffffffff;
	unsigned int tclk = 125000000;

	DP(puts("i2c_init\n"));
	
	REG_READ(CMIC_CONFIG, &control);
	control |= 0x800;
	REG_WRITE(CMIC_CONFIG, control);

	for(n = 0 ; n < 8 ; n++)
	{
		for(m = 0 ; m < 16 ; m++)
		{
			power = 2<<n; /* power = 2^(n+1) */
			freq = tclk/(10*(m+1)*power);
			if (speed > freq)
				margin = speed - freq;
			else
				margin = freq - speed;
			if(margin < minMargin)
			{
				minMargin   = margin;
				actualFreq  = freq;
				actualN	    = n;
				actualM	    = m;
			}
		}
	}

	DP(puts("setup i2c bus\n"));

	/* Setup bus */

	REG_WRITE(I2C_SOFT_RESET, 0);

	DP(puts("udelay...\n"));

	udelay(I2C_DELAY);

	DP(puts("set baudrate\n"));

	REG_WRITE(I2C_STATUS_BAUD_RATE, (actualM << 3) | actualN);
	REG_WRITE(I2C_CONTROL, (0x1 << 2) | (0x1 << 6));

	udelay(I2C_DELAY * 10);

	DP(puts("read control, baudrate\n"));

	REG_READ(I2C_STATUS_BAUD_RATE, &status);
	REG_READ(I2C_CONTROL, &control);
}

static uchar i2c_start(void)
{
	unsigned int control, status;
	int count = 0;

	DP(puts("i2c_start\n"));

	/* Set the start bit */

	REG_READ(I2C_CONTROL, &control);
	control |= (0x1 << 5);
	REG_WRITE(I2C_CONTROL, control);

	REG_READ(I2C_STATUS_BAUD_RATE, &status);

	count = 0;
	while ((status & 0xff) != 0x08) {
		udelay(I2C_DELAY);
		if (count > 20) {
			REG_WRITE(I2C_CONTROL, (0x1 << 4)); /*stop*/
			return (status);
		}
		REG_READ(I2C_STATUS_BAUD_RATE, &status);
		count++;
	}

	return (0);
}

static uchar i2c_select_device(uchar dev_addr, uchar read, int alen)
{
	unsigned int status, data, bits = 7;
	int count = 0;

	DP(puts("i2c_select_device\n"));

	/* Output slave address */

	if (alen > 1) {
		bits = 10;
	}

	data = (dev_addr << 1);
	/* set the read bit */
	data |= read;
	REG_WRITE(I2C_DATA, data);
	/* assert the address */
	RESET_REG_BITS(I2C_CONTROL, BIT3);

	udelay(I2C_DELAY);

	REG_READ(I2C_STATUS_BAUD_RATE, &status);
	count = 0;
	while (((status & 0xff) != 0x40) && ((status & 0xff) != 0x18)) {
		udelay(I2C_DELAY);
		if (count > 20) {
			REG_WRITE(I2C_CONTROL, (0x1 << 4)); /*stop*/
			return(status);
		}
		REG_READ(I2C_STATUS_BAUD_RATE, &status);
		count++;
	}

	if (bits == 10) {
		printf("10 bit I2C addressing not yet implemented\n");
		return (0xff);
	}

	return (0);
}

static uchar i2c_get_data(uchar* return_data, int len) {

	unsigned int data, status = 0;
	int count = 0;

	DP(puts("i2c_get_data\n"));

	while (len) {

		/* Get and return the data */

		RESET_REG_BITS(I2C_CONTROL, (0x1 << 3));

		udelay(I2C_DELAY * 5);

		REG_READ(I2C_STATUS_BAUD_RATE, &status);
		count++;
		while ((status & 0xff) != 0x50) {
			udelay(I2C_DELAY);
			if(count > 2) {
				REG_WRITE(I2C_CONTROL, (0x1 << 4)); /*stop*/
				return 0;
			}
			REG_READ(I2C_STATUS_BAUD_RATE, &status);
			count++;
		}
		REG_READ(I2C_DATA, &data);
		len--;
		*return_data = (uchar)data;
		return_data++;
	}
	RESET_REG_BITS(I2C_CONTROL, BIT2|BIT3);
	while ((status & 0xff) != 0x58) {
		udelay(I2C_DELAY);
		if(count > 200) {
			REG_WRITE(I2C_CONTROL, (0x1 << 4)); /*stop*/
			return (status);
		}
		REG_READ(I2C_STATUS_BAUD_RATE, &status);
		count++;
	}
	REG_WRITE(I2C_CONTROL, (0x1 << 4)); /* stop */

	return (0);
}

static uchar i2c_write_data(unsigned int data, int len)
{
	unsigned int status;
	int count = 0;

	DP(puts("i2c_write_data\n"));

	if (len > 4)
		return -1;

	while (len) {
		/* Set and assert the data */

		REG_WRITE(I2C_DATA, (unsigned int)data);
		RESET_REG_BITS(I2C_CONTROL, (0x1 << 3));

		udelay(I2C_DELAY);

		REG_READ(I2C_STATUS_BAUD_RATE, &status);
		count++;
		while ((status & 0xff) != 0x28) {
			udelay(I2C_DELAY);
			if(count > 20) {
				REG_WRITE(I2C_CONTROL, (0x1 << 4)); /*stop*/
				return (status);
			}
			REG_READ(I2C_STATUS_BAUD_RATE, &status);
			count++;
		}
		len--;
	}
	
	return (0);
}

static void i2c_stop(void)
{
	REG_WRITE(I2C_CONTROL, (0x1 << 3) | (0x1 << 4));
	REG_WRITE(I2C_CONTROL, (0x1 << 4));

	udelay(I2C_DELAY * 10);
}


static uchar i2c_set_dev_offset(uchar dev_addr, unsigned int offset, int alen)
{
	uchar status;

	DP(puts("i2c_set_dev_offset\n"));

	status = i2c_select_device(dev_addr, 0, alen);
	if (status) {
#ifdef DEBUG_I2C
		printf("Failed to select device setting offset: 0x%02x\n",
		       status);
#endif
		return status;
	}

	status = i2c_write_data(offset, 1);
	if (status) {
#ifdef DEBUG_I2C
		printf("Failed to write data: 0x%02x\n", status);
#endif
		return status;
	}

	return (0);
}

int i2c_read(u8 dev_addr, uint offset, int alen, uchar* data, int len)
{
	uchar status = 0;

	DP(puts("i2c_read\n"));

	i2c_init(CFG_I2C_SPEED, 0);

	status = i2c_start();

	if (status) {
#ifdef DEBUG_I2C
		printf("Transaction start failed: 0x%02x\n", status);
#endif
		return status;
	}

	status = i2c_set_dev_offset(dev_addr, offset, 0);
	if (status) {
#ifdef DEBUG_I2C
		printf("Failed to set offset: 0x%02x\n", status);
#endif
		return status;
	}
	
	i2c_stop();

	i2c_init(CFG_I2C_SPEED, 0);

	status = i2c_start();
	if (status) {
#ifdef DEBUG_I2C
		printf("Transaction restart failed: 0x%02x\n", status);
#endif
		return status;
	}

	status = i2c_select_device(dev_addr, 1, alen);
	if (status) {
#ifdef DEBUG_I2C
		printf("Address not acknowledged: 0x%02x\n", status);
#endif
		return status;
	}

	status = i2c_get_data(data, len);
	if (status) {
#ifdef DEBUG_I2C
		printf("Data not recieved: 0x%02x\n", status);
#endif
		return status;
	}

	return 0;
}

int i2c_write (u8 dev_addr, uint offset, int alen, u8 *data, int length)
{
	uchar status = 0;
	int i;

	DP(puts("i2c_write\n"));

	i2c_init(CFG_I2C_SPEED, 0);

	status = i2c_start();
	if (status) {
#ifdef DEBUG_I2C
		printf("Transaction start failed: 0x%02x\n", status);
#endif
		return status;
	}

	status = i2c_set_dev_offset(dev_addr, offset, 0);
	if (status) {
#ifdef DEBUG_I2C
		printf("Failed to set offset: 0x%02x\n", status);
#endif
		return status;
	}

	for (i = 0; i < length; i++) {
		status = i2c_write_data((unsigned int)data[i], 1);
		if (status) {
#ifdef DEBUG_I2C
			printf("Data not written: 0x%02x\n", status);
#endif
			break;
		}
	}
	
	i2c_stop();
	
	return !(i == length);
}

int i2c_probe (uchar chip)
{
	int tmp;

	/*
	 * Try to read the first location of the chip.  The underlying
	 * driver doesn't appear to support sending just the chip address
	 * and looking for an <ACK> back.
	 */
	udelay(10000);
	return i2c_read (chip, 0, 1, (uchar *)&tmp, 1);
}

uchar i2c_reg_read (uchar i2c_addr, uchar reg)
{
	uchar buf[1];

	i2c_read (i2c_addr, reg, 1, buf, 1);

	return (buf[0]);
}

void i2c_reg_write (uchar i2c_addr, uchar reg, uchar val)
{
	i2c_write (i2c_addr, reg, 1, &val, 1);
}
